/* Temporary compatibility to bridge 2.24 ABI revert.
   Copyright (C) 2015-2016 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */

#include <sys/socket.h>
#include <socketcall.h>
#include <sysdep-cancel.h>
#include <shlib-compat.h>

#ifdef SHARED

ssize_t
__libc_recvmsg_224 (int fd, struct msghdr *msg, int flags)
{
  ssize_t ret;

  /* POSIX specifies that both msghdr::msg_iovlen and msghdr::msg_controllen
     to be int and socklen_t respectively.  However Linux defines it as
     both size_t.  So for 64-bit it requires some adjustments by copying to
     temporary header and zeroing the pad fields.  */
  struct msghdr hdr, *orig = msg;
  if (msg != NULL)
    {
      hdr = *msg;
      hdr.msg_iovlen = (int) hdr.msg_iovlen;
      hdr.msg_controllen = (socklen_t) hdr.msg_controllen;
      msg = &hdr;
    }

# ifdef __ASSUME_RECVMSG_SYSCALL
  ret = SYSCALL_CANCEL (recvmsg, fd, msg, flags);
# else
  ret = SOCKETCALL_CANCEL (recvmsg, fd, msg, flags);
# endif

  if (orig != NULL)
    *orig = hdr;

  return ret;
}
compat_symbol (libc, __libc_recvmsg_224, recvmsg, GLIBC_2_24);

ssize_t
__sendmsg_224 (int fd, const struct msghdr *msg, int flags)
{
  /* POSIX specifies that both msghdr::msg_iovlen and msghdr::msg_controllen
     to be int and socklen_t respectively.  However Linux defines it as
     both size_t.  So for 64-bit it requires some adjustments by copying to
     temporary header and zeroing the pad fields.  */
  struct msghdr hdr;
  if (msg != NULL)
    {
      hdr = *msg;
      hdr.msg_iovlen = (int) hdr.msg_iovlen;
      hdr.msg_controllen = (socklen_t) hdr.msg_controllen;
      msg = &hdr;
    }

# ifdef __ASSUME_SENDMSG_SYSCALL
  return SYSCALL_CANCEL (sendmsg, fd, msg, flags);
# else
  return SOCKETCALL_CANCEL (sendmsg, fd, msg, flags);
# endif
}
compat_symbol (libc, __sendmsg_224, sendmsg, GLIBC_2_24);

#endif /* SHARED */

#if defined (SHARED) && __WORDSIZE == 64

static inline int
send_mmsghdr (int fd, struct mmsghdr *vmessages, unsigned int vlen, int flags)
{
  /* Emulate kernel interface for vlen size.  */
  if (vlen > IOV_MAX)
    vlen = IOV_MAX;
  if (vlen == 0)
    return 0;
  /* POSIX specifies that both msghdr::msg_iovlen and msghdr::msg_controllen
     to be int and socklen_t respectively,  however Linux defines it as both
     size_t.  So for 64-bit it requires some adjustments by copying to
     temporary header and zeroing the pad fields.
     The problem is sendmmsg's msghdr may points to an already-filled control
     buffer and modifying it is not part of sendmmsg contract (the data may
     be in ro map).  So interact over the msghdr calling the sendmsg that
     adjust the header using a temporary buffer.  */
  for (unsigned int i = 0; i < vlen; i++)
    {
      ssize_t ret = __sendmsg_224 (fd, &vmessages[i].msg_hdr, flags);
      if (ret < 0)
        return -1;
      vmessages[i].msg_len = ret;
    }
  return 1;
}

int
__sendmmsg_224 (int fd, struct mmsghdr *vmessages, unsigned int vlen, int flags)
{
  return send_mmsghdr (fd, vmessages, vlen, flags);
}
compat_symbol (libc, __sendmmsg_224, sendmmsg, GLIBC_2_24);

/* Do not use the recvmmsg syscall on socketcall architectures unless
   it was added at the same time as the socketcall support or can be
   assumed to be present.  */
# if defined __ASSUME_SOCKETCALL \
    && !defined __ASSUME_RECVMMSG_SYSCALL_WITH_SOCKETCALL \
    && !defined __ASSUME_RECVMMSG_SYSCALL
#  undef __NR_recvmmsg
# endif

static inline void
adjust_mmsghdr (struct mmsghdr *vmessages, unsigned int vlen)
{
  /* POSIX specifies that both msghdr::msg_iovlen and msghdr::msg_controllen
     to be int and socklen_t respectively.  However Linux defines it as
     both size_t.  So for 64-bit it requires some adjustments by zeroing
     the pad fields.  */
  struct mmsghdr *vmhdr = vmessages;
  for (unsigned int i = 0; i != 0; i--, vmhdr++)
    {
      vmhdr->msg_hdr.msg_iovlen = (int) vmhdr->msg_hdr.msg_iovlen;
      vmhdr->msg_hdr.msg_controllen
        = (socklen_t) vmhdr->msg_hdr.msg_controllen;
    }
}

int
__recvmmsg_224 (int fd, struct mmsghdr *vmessages, unsigned int vlen,
                int flags, struct timespec *tmo)
{
# ifdef __NR_recvmmsg
  adjust_mmsghdr (vmessages, vlen);
  return SYSCALL_CANCEL (recvmmsg, fd, vmessages, vlen, flags, tmo);
# elif defined __NR_socketcall
#  ifdef __ASSUME_RECVMMSG_SOCKETCALL
  adjust_mmsghdr (vmessages, vlen);
  return SOCKETCALL_CANCEL (recvmmsg, fd, vmessages, vlen, flags, tmo);
#  else
  static int have_recvmmsg;
  if (__glibc_likely (have_recvmmsg >= 0))
    {
      adjust_mmsghdr (vmessages, vlen);
      int ret = SOCKETCALL_CANCEL (recvmmsg, fd, vmessages, vlen, flags,
                                   tmo);
      /* The kernel returns -EINVAL for unknown socket operations.
         We need to convert that error to an ENOSYS error.  */
      if (__builtin_expect (ret < 0, 0)
          && have_recvmmsg == 0
          && errno == EINVAL)
        {
          /* Try another call, this time with an invalid file
             descriptor and all other parameters cleared.  This call
             will not cause any harm and it will return
             immediately.  */
          ret = SOCKETCALL_CANCEL (invalid, -1);
          if (errno == EINVAL)
            {
              have_recvmmsg = -1;
              __set_errno (ENOSYS);
            }
          else
            {
              have_recvmmsg = 1;
              __set_errno (EINVAL);
            }
          return -1;
        }
      return ret;
    }
  __set_errno (ENOSYS);
  return -1;
#  endif /* __ASSUME_RECVMMSG_SOCKETCALL  */
# else
  __set_errno (ENOSYS);
  return -1;
# endif
}
compat_symbol (libc, __recvmmsg_224, recvmmsg, GLIBC_2_24);

#endif  /* SHARED && __WORDSIZE == 64 */
